/* blest2exons.c -- convert blest output to an exons file
 *
 * $Revision: 1.17 $$Date: 2001/05/07 18:35:20 $
 */

#include "util.h"

#define SYNTAX  "blest_file [-min min_exons] > output_file"

enum { BUFSIZE = 1000 };

/* typedefs for list of ESTs, each with a header and an exon/intron list */

typedef struct Exon {
	int from, to;
	struct Exon *next_exon;
} *Exon_ptr;

typedef struct Est {
	char *header;
	int from, to, dir, nexon;
	Exon_ptr exons;
	struct Est *next_est;
} *Est_ptr;

static void get_args (int argc, char **argv, char **filename, int *min_exons);
static int est_compare (const void *e, const void *f);

int main (int argc, char **argv) {
	char *filename = NULL;                  /* name of blest file           */
	int min_exons = 1;
	char buf [BUFSIZE], *p;
	FILE *fp;
	Est_ptr est, new_est, Est_array, est_list;
	Exon_ptr exon, new_exon;
	int i, nbr_est, last_exon;

	/* get arguments */
	set_argv0 (argv[0]);
	get_args (argc, argv, &filename, &min_exons);
	if (!filename)
		fatal (usage (SYNTAX));
	if (min_exons <= 0)
		fatal ("min_exons must be positive.");

	/* read blest file into linked list */
	fp = (strsame (filename, "-") ? stdin : ckopen (filename, "r"));
	est_list = NULL;
	for (nbr_est = 0; fgets (buf, BUFSIZE, fp); nbr_est++) {
		if (buf[0] != '>')
			fatalf ("?>?:\n%s", buf);
		new_est = (Est_ptr) ckalloc (sizeof (*est));
		new_est->next_est = est_list;
		est_list = new_est;
		new_est->header = copy_string (buf);
		new_est->from = new_est->to = new_est->nexon = 0;
		new_est->exons = NULL;
		if (!fgets (buf, BUFSIZE, fp))
			fatal ("Unexpected EOF; cannot find EST length.");
		new_est->dir = (strstr (buf, "complement") ? '>' : '<');
		exon = NULL;
		for ( ; ; ) {
			if (!fgets (buf, BUFSIZE, fp))
				fatal ("Unexpected EOF.");
			new_est->nexon++;
			new_exon = (Exon_ptr) ckalloc (sizeof (*exon));
			if (sscanf (buf, "%d-%d",
			    &(new_exon->from), &(new_exon->to)) != 2)
				fatalf ("Cannot find from/to:\n%s", buf);
			if (!new_est->from)
				new_est->from = new_exon->from;
			new_est->to = new_exon->to;
			last_exon = 0;
			if (!(p = strchr (buf, '%')))
				fatalf ("Cannot find '%':\n%s", buf);
			if (!strncmp (p+2, "->", 2))
				new_est->dir = '>';
			else if (!strncmp (p+2, "<-", 2))
				new_est->dir = '<';
			else if (strncmp (p+2, "--", 2) && strncmp (p+2, "==", 2))
				last_exon = 1;
			if (!new_est->exons)
				exon = new_est->exons = new_exon;
			else
				exon = exon->next_exon = new_exon;
			if (last_exon)
				break;
		}
		exon->next_exon = NULL;
	}
	if (!nbr_est)
		fatal ("No EST hits.");

	/* sort ESTs by start position */
	Est_array = (Est_ptr) ckalloc (nbr_est * sizeof (*est));
	for (i = 0, est = est_list; est; i++, est = est->next_est)
		Est_array[i] = *est;
	qsort ((void *) Est_array, nbr_est, sizeof (*est), est_compare);

	/* print the sorted ESTs */
	for (i = 0; i < nbr_est; i++) {
		if (Est_array[i].nexon < min_exons)
			continue;
		if ((p = strchr (Est_array[i].header, '\n')))
			*p = '\0';
		if ((p = strchr (Est_array[i].header, ' ')))
			*p = '\0';
		printf ("\n%c %d %d %s\n",
			Est_array[i].dir,
			Est_array[i].from,
			Est_array[i].to,
			Est_array[i].header+1);
		for (exon = Est_array[i].exons; exon; exon = exon->next_exon)
			printf ("%d %d\n", exon->from, exon->to);
	}

	/* clean up */
	for (est = est_list; est; est = new_est) {
		new_est = est->next_est;
		free (est->header);
		for (exon = est->exons; exon; exon = new_exon) {
			new_exon = exon->next_exon;
			free (exon);
		}
		free (est);
	}
	free (Est_array);
	if (fp != stdin)
		fclose (fp);
	return 0;
}

static void get_args (int argc, char **argv, char **filename, int *min_exons) {
	int i;

	for (i = 1; i < argc; i++) {
		if (strsame (argv[i], "-min")) {
			if (!(*min_exons = atoi (optval (argc, argv, &i, SYNTAX))))
				fatal ("Invalid min_exons value.");
		}
		else if (!*filename)
			*filename = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}

static int est_compare (const void *e, const void *f) {
	int e_from = ((const struct Est *) e)->from;
	int f_from = ((const struct Est *) f)->from;
	return (e_from < f_from ? -1 : 1);
}
